# Copyright 2020, Charles Powell

import hashlib
from time import time

class PlugInstance:

    def __init__(self, id, start_time=None, alias=None, power=0, current=0, voltage=120, mac=None, device_id=None):      
        self.id = id
        
        self.voltage = voltage
        self.power = power
        self.current = current
        
        if alias:
            self.alias = alias
        else:
            self.alias = id
            
        if start_time: self.start_time = start_time
        else: self.start_time = time()-1
        
        if not self.power:
            self.power = self.voltage*self.current
        if not self.current:
            self.current = self.power/self.voltage

        if device_id: self.device_id = device_id
        else: self.device_id = self.generate_deviceid()
        
        if mac: self.mac = mac
        else: self.mac = self.generate_mac()

    def generate_mac(self):
        end = [int(self.device_id[i:i+2],16) for i in range(0, len(self.device_id), 2)]
        mac = [53, 75, 31] + end[:3]
        return ':'.join('%02x' % b for b in mac)

    def generate_deviceid(self):
        return hashlib.sha1(self.id.encode('utf-8')).hexdigest()

    def generate_response(self):
        # Response dict
        response = {
            "emeter": {
                "get_realtime": {
                    "current": self.current,
                    "voltage": self.voltage,
                    "power": self.power,
                    "total": 0,     # Unsure if this needs a value, appears not to
                    "err_code": 0   # No errors here!
                }
            },
            "system": {
                "get_sysinfo": {
                    "err_code": 0,
                    "sw_ver": "1.2.5 Build 171206 Rel.085954",
                    "hw_ver": "1.0",
                    "type": "IOT.SMARTPLUGSWITCH",
                    "model": "HS110(US)",
                    "mac": self.mac.upper(),
                    "deviceId": self.device_id.upper(),
                    "hwId": "60FF6B258734EA6880E186F8C96DDC61",
                    "fwId": "00000000000000000000000000000000",
                    "oemId": "FFF22CFF774A0B89F7624BFC6F50D5DE",
                    "alias": self.alias,
                    "dev_name": "Wi-Fi Smart Plug With Energy Monitoring",
                    "icon_hash": "",
                    "relay_state": 1 if self.power > 0 else 0,
                    "on_time": time() - self.start_time,
                    "active_mode": "none",
                    "feature": "TIM:ENE",
                    "updating": 0,
                    "rssi": -60,    # Great wifi signal
                    "led_off": 0,   # Probably not important
                    "latitude": 39.8283,    # Center of the US
                    "longitude": -98.5795   # Center of the US
                }
            }
        }
        return response
 



import asyncio
import json
import ssl

import aiohttp
import websockets

from .sense_api import *
from .sense_exceptions import *


class ASyncSenseable(SenseableBase):
    def __init__(
        self,
        username=None,
        password=None,
        api_timeout=API_TIMEOUT,
        wss_timeout=WSS_TIMEOUT,
        client_session=None,
    ):
        """Init the ASyncSenseable object."""
        self._client_session = client_session or aiohttp.ClientSession()
        self.ssl_context = None

        super().__init__(
            username=username,
            password=password,
            api_timeout=api_timeout,
            wss_timeout=wss_timeout,
        )

    async def authenticate(self, username, password, ssl_verify=True, ssl_cafile=""):
        auth_data = {"email": username, "password": password}

        # Use custom ssl verification, if specified
        if not ssl_verify:
            self.ssl_context = ssl.create_default_context()
            self.ssl_context.check_hostname = False
            self.ssl_context.verify_mode = ssl.CERT_NONE
        elif ssl_cafile:
            self.ssl_context = ssl.create_default_context(cafile=ssl_cafile)
        else:
            self.ssl_context = ssl.create_default_context()

        # Get auth token
        try:
            async with self._client_session.post(
                API_URL + "authenticate", timeout=self.api_timeout, data=auth_data
            ) as resp:

                # check for 200 return
                if resp.status != 200:
                    raise SenseAuthenticationException(
                        "Please check username and password. API Return Code: %s"
                        % resp.status
                    )

                # Build out some common variables
                self.set_auth_data(await resp.json())
        except Exception as ex:
            raise SenseAPITimeoutException("Connection failure: %s" % ex) from ex

    # Update the realtime data for asyncio
    async def update_realtime(self):
        # rate limit API calls
        if (
            self._realtime
            and self.rate_limit
            and self.last_realtime_call + self.rate_limit > time()
        ):
            return self._realtime
        self.last_realtime_call = time()
        await self.async_realtime_stream(single=True)

    async def async_realtime_stream(self, callback=None, single=False):
        """ Reads realtime data from websocket"""
        url = WS_URL % (self.sense_monitor_id, self.sense_access_token)
        # hello, features, [updates,] data
        async with websockets.connect(url, ssl=self.ssl_context) as ws:
            while True:
                try:
                    message = await asyncio.wait_for(
                        ws.recv(), timeout=self.wss_timeout
                    )
                except asyncio.TimeoutError as ex:
                    raise SenseAPITimeoutException("API websocket timed out") from ex

                result = json.loads(message)
                if result.get("type") == "realtime_update":
                    data = result["payload"]
                    self.set_realtime(data)
                    if callback:
                        callback(data)
                    if single:
                        return
                elif result.get("type") == "error":
                    data = result["payload"]
                    raise SenseWebsocketException(data["error_reason"])

    async def get_realtime_future(self, callback):
        """ Returns an async Future to parse realtime data with callback"""
        await self.async_realtime_stream(callback)

    async def api_call(self, url, payload={}):
        timeout = aiohttp.ClientTimeout(total=self.api_timeout)
        try:
            async with self._client_session.get(
                API_URL + url, headers=self.headers, timeout=timeout, data=payload
            ) as resp:
                return await resp.json()
        except asyncio.TimeoutError as ex:
            # timed out
            raise SenseAPITimeoutException("API call timed out") from ex

    async def get_trend_data(self, scale):
        if scale.upper() not in valid_scales:
            raise Exception("%s not a valid scale" % scale)
        t = datetime.now().replace(hour=12)
        json = self.api_call(
            "app/history/trends?monitor_id=%s&scale=%s&start=%s"
            % (self.sense_monitor_id, scale, t.isoformat())
        )
        self._trend_data[scale] = await json

    async def update_trend_data(self):
        for scale in valid_scales:
            await self.get_trend_data(scale)

    async def get_discovered_device_names(self):
        # lots more info in here to be parsed out
        json = self.api_call("app/monitors/%s/devices" % self.sense_monitor_id)
        self._devices = await [entry["name"] for entry in json]
        return self._devices

    async def get_discovered_device_data(self):
        json = self.api_call("monitors/%s/devices" % self.sense_monitor_id)
        return await json



# Based on: https://github.com/softScheck/tplink-smartplug/blob/dcf978b970356c3edd941583d277612182381f2c/tplink_smartplug.py
#
# TP-Link Wi-Fi Smart Plug Protocol Client
# For use with TP-Link HS-100 or HS-110
#
# by Lubomir Stroetmann
# Copyright 2016 softScheck GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Modifications Copyright (C) 2020, Charles Powell

from struct import pack


def _generate_tplink(unencrypted):
    key = 171
    for unencryptedbyte in unencrypted:
        key = key ^ unencryptedbyte
        yield key


def tp_link_encrypt(string):
    unencrypted = string.encode()
    return pack(">I", len(unencrypted)) + bytes(_generate_tplink(unencrypted))


def tp_link_decrypt(string):
    key = 171
    result = ""
    for i in string:
        a = key ^ i
        key = i
        result += chr(a)
    return result



# Copyright 2020, Charles Powell

# SenseLink is a tool that emulates the energy monitoring functionality of TP-Link Kasa HS110 Smart Plugs, 
# and allows you to report "custom" power usage to your Sense Home Energy Monitor based on other parameters. 

import logging
import asyncio
import json

from .tplink_encryption import *
from .plug_instance import PlugInstance

SENSE_TP_LINK_PORT = 9999

class SenseLinkServerProtocol:
    def __init__(self, devices):
        self._devices = devices
        self.should_respond = True
        
    def connection_made(self, transport):
        self.transport = transport
        
    def connection_lost(self, exc):
        pass

    def datagram_received(self, data, addr):
        decrypted_data = tp_link_decrypt(data)

        try:
            json_data = json.loads(decrypted_data)
            # Sense requests the emeter and system parameters
            if 'emeter' in json_data and 'get_realtime' in json_data["emeter"] and \
               'system' in json_data and 'get_sysinfo' in json_data["system"]:
                # Check for non-empty values, to prevent echo storms
                if json_data['emeter']['get_realtime']:
                    # This is a self-echo, common with Docker without --net=Host!
                    logging.debug("Ignoring non-empty/non-Sense UDP request")
                    return

                logging.debug("Broadcast received from: %s: %s", addr, json_data)

                # Build and send responses
                for plug in self._devices():
                    # Build response
                    response = plug.generate_response()
                    json_resp = json.dumps(response, separators=(',', ':'))
                    encrypted_resp = tp_link_encrypt(json_resp)
                    # Strip leading 4 bytes for...some reason
                    encrypted_resp = encrypted_resp[4:]

                    # Allow disabling response
                    if self.should_respond:
                        # Send response
                        logging.debug(f"Sending response: {response}")
                        self.transport.sendto(encrypted_resp, addr)
                    else:
                        # Do not send response, but log for debugging
                        logging.debug(f"SENSE_RESPONSE disabled, response content: {response}")
            else:
                logging.info(f"Unexpected/unhandled message: {json_data}")

        # Appears to not be JSON
        except ValueError:
            logging.debug("Did not receive valid json")


class SenseLink:
    _devices = []

    def __init__(self, devices, port=SENSE_TP_LINK_PORT):
        self.port = port
        self._devices = devices

    def print_instance_wattages(self):
        for inst in self._devices():
            logging.info(f"Plug {inst.alias} power: {inst.power}")

    async def start(self):
        loop = asyncio.get_running_loop()
        self.transport, self.protocol = await loop.create_datagram_endpoint(
            lambda: SenseLinkServerProtocol(self._devices),
            local_addr=('0.0.0.0', self.port))
            
    async def stop(self):
        self.transport.close()



import json
import requests
import ssl
from requests.exceptions import ReadTimeout
from websocket import create_connection
from websocket._exceptions import WebSocketTimeoutException

from .sense_api import *
from .sense_exceptions import *

class Senseable(SenseableBase):

    def authenticate(self, username, password, ssl_verify=True, ssl_cafile=''):
        auth_data = {
            "email": username,
            "password": password
        }
        
        # Create session
        self.s = requests.session()

        # Use custom ssl verification, if specified
        if not ssl_verify:
            self.s.verify = False
        elif ssl_cafile:
            self.s.verify = ssl_cafile

        # Get auth token
        try:
            response = self.s.post(API_URL+'authenticate',
                                   auth_data, timeout=self.api_timeout)
        except Exception as e:
            raise Exception('Connection failure: %s' % e)

        # check for 200 return
        if response.status_code != 200:
            raise SenseAuthenticationException(
                "Please check username and password. API Return Code: %s" %
                response.status_code)
        
        self.set_auth_data(response.json())
    
    # Update the realtime data 
    def update_realtime(self):
        # rate limit API calls
        if self._realtime and self.rate_limit and \
           self.last_realtime_call + self.rate_limit > time():
            return self._realtime
        url = WS_URL % (self.sense_monitor_id, self.sense_access_token)
        next(self.get_realtime_stream())
    
    def get_realtime_stream(self):
        """ Reads realtime data from websocket
            Continues until loop broken"""
        ws = 0
        url = WS_URL % (self.sense_monitor_id, self.sense_access_token)
        try:
            ws = create_connection(url, timeout=self.wss_timeout, sslopt={"cert_reqs": ssl.CERT_NONE})
            while True: # hello, features, [updates,] data
                result = json.loads(ws.recv())
                if result.get('type') == 'realtime_update':
                    data = result['payload']
                    self.set_realtime(data)
                    yield data
        except WebSocketTimeoutException:
            raise SenseAPITimeoutException("API websocket timed out")
        finally:
            if ws: ws.close()    
    
    def get_trend_data(self, scale):
        if scale.upper() not in valid_scales:
            raise Exception("%s not a valid scale" % scale)
        t = datetime.now().replace(hour=12)
        self._trend_data[scale] = self.api_call(
            'app/history/trends?monitor_id=%s&scale=%s&start=%s' %
            (self.sense_monitor_id, scale, t.isoformat()))

    def update_trend_data(self):
        for scale in valid_scales:
            self.get_trend_data(scale)

    def api_call(self, url, payload={}):
        try:
            return self.s.get(API_URL + url,
                              headers=self.headers,
                              timeout=self.api_timeout,
                              data=payload).json()
        except ReadTimeout:
            raise SenseAPITimeoutException("API call timed out")   

    def get_discovered_device_names(self):
        # lots more info in here to be parsed out
        json = self.api_call('app/monitors/%s/devices' %
                                 self.sense_monitor_id)
        self._devices = [entry['name'] for entry in json]
        return self._devices

    def get_discovered_device_data(self):
        return self.api_call('monitors/%s/devices' %
                             self.sense_monitor_id)

    def always_on_info(self):
        # Always on info - pretty generic similar to the web page
        return self.api_call('app/monitors/%s/devices/always_on' %
                             self.sense_monitor_id)

    def get_monitor_info(self):
        # View info on your monitor & device detection status
        return self.api_call('app/monitors/%s/status' %
                             self.sense_monitor_id)

    def get_device_info(self, device_id):
        # Get specific informaton about a device
        return self.api_call('app/monitors/%s/devices/%s' %
                             (self.sense_monitor_id, device_id))

    def get_all_usage_data(self):
        payload = {'n_items': 30}
        # lots of info in here to be parsed out
        return self.api_call('users/%s/timeline' %
                          (self.sense_user_id), payload)





from .sense_api import SenseableBase
from .sense_exceptions import *

from .senseable import Senseable
import sys
if sys.version_info >= (3, 5):
    from .asyncsenseable import ASyncSenseable
    from .plug_instance import PlugInstance
    from .sense_link import SenseLink

__version__ = "0.9.2"



import json
import sys
from time import time
from datetime import datetime

from .sense_exceptions import *

API_URL = 'https://api.sense.com/apiservice/api/v1/'
WS_URL = "wss://clientrt.sense.com/monitors/%s/realtimefeed?access_token=%s"
API_TIMEOUT = 5
WSS_TIMEOUT = 5
RATE_LIMIT = 60

# for the last hour, day, week, month, or year
valid_scales = ['HOUR', 'DAY', 'WEEK', 'MONTH', 'YEAR']


class SenseableBase(object):

    def __init__(self, username=None, password=None,
                 api_timeout=API_TIMEOUT, wss_timeout=WSS_TIMEOUT):

        # Timeout instance variables
        self.api_timeout = api_timeout
        self.wss_timeout = wss_timeout
        self.rate_limit = RATE_LIMIT
        self.last_realtime_call = 0

        self._realtime = {}
        self._devices = []
        self._trend_data = {}
        for scale in valid_scales: self._trend_data[scale] = {}

        if username and password:
            self.authenticate(username, password)

    def set_auth_data(self, data):
        self.sense_access_token = data['access_token']
        self.sense_user_id = data['user_id']
        self.sense_monitor_id = data['monitors'][0]['id']

        # create the auth header
        self.headers = {'Authorization': 'bearer {}'.format(
            self.sense_access_token)}

    @property
    def devices(self):
        """Return devices."""
        return self._devices

    def set_realtime(self, data):
        self._realtime = data
        self.last_realtime_call = time()

    def get_realtime(self):
        return self._realtime

    @property
    def active_power(self):
        return self._realtime.get('w', 0)

    @property
    def active_solar_power(self):
        return self._realtime.get('solar_w', 0)

    @property
    def active_voltage(self):
        return self._realtime.get('voltage', [])

    @property
    def active_frequency(self):
        return self._realtime.get('hz', 0)

    @property
    def daily_usage(self):
        return self.get_trend('DAY', 'consumption')

    @property
    def daily_production(self):
        return self.get_trend('DAY', 'production')

    @property
    def daily_production_pct(self):
        return self.get_trend('DAY', 'production_pct')

    @property
    def daily_net_production(self):
        return self.get_trend('DAY', 'net_production')

    @property
    def daily_from_grid(self):
        return self.get_trend('DAY', 'from_grid')

    @property
    def daily_to_grid(self):
        return self.get_trend('DAY', 'to_grid')

    @property
    def daily_solar_powered(self):
        return self.get_trend('DAY', 'solar_powered')

    @property
    def weekly_usage(self):
        return self.get_trend('WEEK', 'consumption')

    @property
    def weekly_production(self):
        return self.get_trend('WEEK', 'production')

    @property
    def weekly_production_pct(self):
        return self.get_trend('WEEK', 'production_pct')

    @property
    def weekly_net_production(self):
        return self.get_trend('WEEK', 'net_production')

    @property
    def weekly_from_grid(self):
        return self.get_trend('WEEK', 'from_grid')

    @property
    def weekly_to_grid(self):
        return self.get_trend('WEEK', 'to_grid')

    @property
    def weekly_solar_powered(self):
        return self.get_trend('WEEK', 'solar_powered')

    @property
    def monthly_usage(self):
        return self.get_trend('MONTH', 'consumption')

    @property
    def monthly_production(self):
        return self.get_trend('MONTH', 'production')

    @property
    def monthly_production_pct(self):
        return self.get_trend('MONTH', 'production_pct')

    @property
    def monthly_net_production(self):
        return self.get_trend('MONTH', 'net_production')

    @property
    def monthly_from_grid(self):
        return self.get_trend('MONTH', 'from_grid')

    @property
    def monthly_to_grid(self):
        return self.get_trend('MONTH', 'to_grid')

    @property
    def monthly_solar_powered(self):
        return self.get_trend('MONTH', 'solar_powered')

    @property
    def yearly_usage(self):
        return self.get_trend('YEAR', 'consumption')

    @property
    def yearly_production(self):
        return self.get_trend('YEAR', 'production')

    @property
    def yearly_production_pct(self):
        return self.get_trend('YEAR', 'production_pct')

    @property
    def yearly_net_production(self):
        return self.get_trend('YEAR', 'net_production')

    @property
    def yearly_from_grid(self):
        return self.get_trend('YEAR', 'from_grid')

    @property
    def yearly_to_grid(self):
        return self.get_trend('YEAR', 'to_grid')

    @property
    def yearly_solar_powered(self):
        return self.get_trend('YEAR', 'solar_powered')

    @property
    def active_devices(self):
        return [d['name'] for d in self._realtime.get('devices', {})]

    def get_trend(self, scale, key):
        if isinstance(key, bool):
            key = 'production' if key is True else 'consumption'
        else:
            key = 'consumption' if key == 'usage' else key
        if key not in self._trend_data[scale]: return 0
        # Perform a check for a valid type
        if not isinstance(self._trend_data[scale][key], (dict, float, int)): return 0
        if isinstance(self._trend_data[scale][key], dict):
            total = self._trend_data[scale][key].get('total', 0)
        else:
            total = self._trend_data[scale][key]
        return total
